/* A trivial program (should be invoked by inetd) that
 * writes out a string "250 ESMTP_unwelcome [peer.ip.add.ress]"
 * to the network and optionally waits for a string in form
 *    [junk]protocol:ip.add.re.ss:port\n
 * from the remote system.  May be used as a destination for
 * proxycheck program.  All connections (together with the
 * information in the above form, if given) are optionally
 * logged to a specified file.
 * Options:
 *   -n - do not wait for a proxy connection info
 *   -t timeout - timeout in secs
 *   -l logfile - where to log proxy/connection info
 */

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

struct sockaddr_in peer;
char *logfile;

void info(char *ext) {
  int logfd;
  if (logfile && (logfd = open(logfile, O_CREAT|O_WRONLY|O_APPEND, 0644)) > 0) {
    char buf[128];
    time_t t = time(NULL);
    int l = sprintf(buf, "%.24s %s", ctime(&t), inet_ntoa(peer.sin_addr));
    if (ext)
      l += sprintf(buf + l, " %s", ext);
    buf[l++] = '\n';
    write(logfd, buf, l);
    close(logfd);
  }
  exit(0);
}

void timedout(int sig) {
  info(NULL);
}

void findproxyinfo(char *buf, char *be) {
  const char *proxyproto;
  struct in_addr proxyaddr;
  int proxyport;
  int i, l;
  char *p;
  char *b = buf;
  static const char *protos[] = { "socks4", "socks5", "http", "wingate", NULL };

  *be = '\0';

  while(b < be) {
    if (*b++ != '\n') continue;
    for(i = 0; protos[i]; ++i) {
      l = strlen(protos[i]);
      if (b + l < be && b[l] == ':' && memcmp(protos[i], b, l) == 0)
	 break;
    }
    if (!protos[i]) continue;
    proxyproto = protos[i];
    b += l + 1;
    if ((p = strchr(b, ':')) == NULL) continue;
    *p = '\0';
    proxyaddr.s_addr = inet_addr(b);
    *p++ = ':';
    if (proxyaddr.s_addr == INADDR_NONE) continue;
    b = p;
    proxyport = 0;
    while(*b >= '0' && *b <= '9' && b - p < 6)
      proxyport = proxyport * 10 + (*b++ - '0');
    if (*b == '\n' || *b == '\r' || *b == ' ') {
      sprintf(buf, "%s:%s:%d", proxyproto, inet_ntoa(proxyaddr), proxyport);
      info(buf);
    }
  }
}

int main(int argc, char **argv) {
  int timeout = 30;
  char buf[2048];
  int nowait = 0;
  int l, r;

  while((l = getopt(argc, argv, "t:l:n")) != EOF)
    switch(l) {
    case 't': timeout = atoi(optarg); break;
    case 'l': logfile = optarg; break;
    case 'n': nowait = 1; break;
    default: return 1;
    }

  l = sizeof(peer);
  if (getpeername(0, (struct sockaddr*)&peer, &l) < 0 ||
      l != sizeof(peer) || peer.sin_family != AF_INET)
    return 1;

  signal(SIGALRM, timedout);
  alarm(timeout);

  /* print some easily recognizeable string */
  l = sprintf(buf, "250 ESMTP_unwelcome [%s]\r\n", inet_ntoa(peer.sin_addr));
  write(0, buf, l);

  if (nowait || !logfile)
    info(NULL);

  /* read data from the network.
   * try to recognize:
   *  junk space proto:ipaddr:port space
   */
  l = 1;
  buf[0] = '\n';
  while((r = read(0, buf + l, sizeof(buf) - l - 1)) > 0) {
    l += r;
    findproxyinfo(buf, buf + l);
  }
  info(NULL);
  return 0;
}
